@@ -8,7 +8,7 @@
  *             Andrae Muys <andrae@humbug.org.au.au>
  *           - uses libbzip2 by Julian Seward (http://sources.redhat.com/bzip2/)
  *
- *	Date  : July 25, 2007
+ *	Date  : October 31, 2008
  *
  *
  *  Contributions
@@ -34,20 +34,28 @@
  *  Andrew Pantyukhin <sat@FreeBSD.org> - maintains the FreeBSD ports and willing to
  *          resolve any FreeBSD-related problems
  *  Roland Illig <rillig@NetBSD.org> - creates and maintains NetBSD packages
+ *  Matt Turner <mattst88@gmail.com> - code cleanup
+ *  lvaro Reguly <alvaro@reguly.com> - RPM spec update to support SUSE Linux
+ *  Ivan Voras <ivoras@freebsd.org> - support for stdin and pipes during compression and
+ *          CPU detect changes
+ *  John Dalton <john@johndalton.info> - code cleanup and bug fixes for stdin support
+ *  Rene Georgi <rene.georgi@online.de> - code and Makefile cleanup, support for direct
+ *          decompress and bzcat
  *
  *  Specials thanks for suggestions and testing:  Phillippe Welsh,
  *  James Terhune, Dru Lemley, Bryan Stillwell, George Chalissery,
  *  Kir Kolyshkin, Madhu Kangara, Mike Furr, Joergen Ramskov, Kurt Fitzner,
  *  Peter Cordes, Oliver Falk, Jindrich Novy, Benjamin Reed, Chris Dearman,
  *  Richard Russon, Anbal Monsalve Salazar, Jim Leonard, Paul Pluzhnikov,
- *  Coran Fisher, Ken Takusagawa, David Pyke.
+ *  Coran Fisher, Ken Takusagawa, David Pyke, Matt Turner, Damien Ancelin,
+ *  lvaro Reguly, Ivan Voras, John Dalton, Sami Liedes, Rene Georgi.
  *
  *
- * This program, "pbzip2" is copyright (C) 2003-2007 Jeff Gilchrist.
+ * This program, "pbzip2" is copyright (C) 2003-2008 Jeff Gilchrist.
  * All rights reserved.
  *
  * The library "libbzip2" which pbzip2 uses, is copyright
- * (C) 1996-2006 Julian R Seward.  All rights reserved.
+ * (C) 1996-2008 Julian R Seward.  All rights reserved.
  *
  * Redistribution and use in source and binary forms, with or without
  * modification, are permitted provided that the following conditions
@@ -82,7 +90,7 @@
  *
  * Jeff Gilchrist, Ottawa, Canada.
  * pbzip2@compression.ca
- * pbzip2 version 1.0.2 of July 25, 2007
+ * pbzip2 version 1.0.3 of October 31, 2008
  *
  */
 #include <vector>
@@ -195,6 +203,7 @@ static int NumBufferedBlocks = 0;
 static int Verbosity = 0;
 static int QuietMode = 1;
 static int OutputStdOut = 0;
+static int ForceOverwrite = 0;
 static int BWTblockSize = 9;
 static int FileListCount = 0;
 static std::vector <outBuff> OutputBuffer;
@@ -212,7 +221,8 @@ int producer_decompress(int, OFF_T, queue *);
 void *consumer_decompress(void *);
 void *fileWriter(void *);
 int directcompress(int, OFF_T, int, char *);
-int producer(int, int, OFF_T, int, queue *);
+int directdecompress(char *, char *);
+int producer(int, int, int, queue *);
 void *consumer(void *);
 queue *queueInit(int);
 void queueDelete(queue *);
@@ -222,6 +232,8 @@ int getFileMetaData(char *);
 int writeFileMetaData(char *);
 int testBZ2ErrorHandling(int, BZFILE *, int);
 int testCompressedData(char *);
+ssize_t bufread(int hf, char *buf, size_t bsize);
+int detectCPUs(void);
 
 
 /*
@@ -271,34 +283,17 @@ void mySignalCatcher(int n)
 char *memstr(char *searchBuf, int searchBufSize, char *searchString, int searchStringSize)
 {
 	int i;
-	int j;
-	int notFound = 0;
 
 	for (i=0; i < searchBufSize; i++)
 	{
 		if ((searchBufSize - i) < searchStringSize)
 			break;
 
-		if (searchBuf[i] == searchString[0])
+		if ( searchBuf[i] == searchString[0] && 
+			 memcmp(searchBuf+i, searchString, searchStringSize) == 0 ) 
 		{
-			notFound = 0;
-			// we found first character now check rest
-			for (j=1; j < searchStringSize; j++)
-			{
-				if (searchBuf[i+j] != searchString[j])
-				{	notFound = 1;
-					break;
-				}
-			}
-			// we found the first occurance of the string so we are done
-			if (notFound == 0)
-			{
-				#ifdef PBZIP_DEBUG
-				fprintf(stderr, " memstr(): Found substring at byte: %d [%x]\n", i, &searchBuf[i]);
-				#endif
-				return &searchBuf[i];
-			}
-		}
+			return &searchBuf[i];
+		}	
 	}
 
 	return NULL;
@@ -432,6 +427,15 @@ int producer_decompress(int hInfile, OFF_T fileSize, queue *fifo)
 		bytesLeft -= ret;
 	} // while
 
+	// use direcdecompress() instead to process 1 bzip2 stream
+	if (bz2NumBlocks <= 1)
+	{
+		if (QuietMode != 1)
+			fprintf(stderr, "Switching to no threads mode: only 1 BZIP2 block found.\n");
+		allDone = 1;
+		return -99;
+	}
+
 	pthread_mutex_lock(MemMutex);
 	if (FileData != NULL)
 		delete [] FileData;
@@ -443,6 +447,7 @@ int producer_decompress(int hInfile, OFF_T fileSize, queue *fifo)
 	if (OutputBuffer.size() != bz2NumBlocks)
 	{
 		fprintf(stderr, " *ERROR: Could not allocate memory (OutputBuffer)!  Aborting...\n");
+		allDone = 1;
 		return 1;
 	}
 	// set empty buffer
@@ -566,7 +571,8 @@ int producer_decompress(int hInfile, OFF_T fileSize, queue *fifo)
 
 		// add data to the decompression queue
 		pthread_mutex_lock (fifo->mut);
-		while (fifo->full) {
+		while (fifo->full) 
+		{
 			#ifdef PBZIP_DEBUG
 			printf ("producer: queue FULL.\n");
 			#endif
@@ -581,6 +587,20 @@ int producer_decompress(int hInfile, OFF_T fileSize, queue *fifo)
 		pthread_cond_signal (fifo->notEmpty);
 
 		blockNum++;
+
+		// make sure output buffer is large enough to handle input data
+		if (blockNum > OutputBuffer.size()) 
+		{
+			int newsize = OutputBuffer.size()*2;
+			pthread_mutex_lock(OutMutex);
+			OutputBuffer.resize(newsize);
+			pthread_mutex_unlock(OutMutex);
+			if (OutputBuffer.size() != newsize)
+			{
+				allDone = 1;
+				return -1;
+			}
+		}
 	} // for
 
 	close(hInfile);
@@ -775,6 +795,10 @@ void *fileWriter(void *outname)
 
 	while ((currBlock < NumBlocks) || (allDone == 0))
 	{
+		#ifdef PBZIP_DEBUG
+		fprintf(stderr, "fileWriter:  Block: %d\n", currBlock);
+		#endif
+
 		pthread_mutex_lock(OutMutex);
 		if ((OutputBuffer.size() == 0) || (OutputBuffer[currBlock].bufSize < 1) || (OutputBuffer[currBlock].buf == NULL))
 		{
@@ -834,7 +858,7 @@ void *fileWriter(void *outname)
 
 	if (OutputStdOut == 0)
 		close(hOutfile);
-	if (QuietMode != 1)
+	if ((QuietMode != 1) && (allDone == 0))
 	{
 		fprintf(stderr, "    Output Size: %"PRIu64" bytes\n", (unsigned long long)CompressedSize);
 	}
@@ -1025,25 +1049,266 @@ int directcompress(int hInfile, OFF_T fileSize, int blockSize, char *OutFilename
 /*
  *********************************************************
  */
-int producer(int hInfile, OFF_T fileSize, int blockSize, queue *fifo)
+int directdecompress(char *InFilename, char *OutFilename)
+{
+	FILE *stream = NULL;
+	FILE *zStream = NULL;
+	BZFILE* bzf = NULL;
+	unsigned char obuf[5000];
+	unsigned char unused[BZ_MAX_UNUSED];
+	unsigned char *unusedTmp;
+	int bzerr, nread, streamNo;
+	int nUnused;
+	int ret = 0;
+	int i;
+
+	nUnused = 0;
+	streamNo = 0;
+
+	// see if we are using stdin or not
+	if (strcmp(InFilename, "-") != 0) 
+	{
+		// open the file for reading
+		zStream = fopen(InFilename, "rb");
+		if (zStream == NULL)
+		{
+			fprintf(stderr, " *ERROR: Could not open input file [%s]!  Skipping...\n", InFilename);
+			return -1;
+		}
+	}
+	else
+		zStream = stdin;
+
+	// check file stream for errors
+	if (ferror(zStream))
+	{
+		fprintf(stderr, " *ERROR: Problem with input stream of file [%s]!  Skipping...\n", InFilename);
+		if (zStream != stdin)
+			fclose(zStream);
+		return -1;
+	}
+
+	// see if we are outputting to stdout
+	if (OutputStdOut == 0)
+	{
+        stream = fopen(OutFilename, "wb");
+	}
+	else
+		stream = stdout;
+
+	// check file stream for errors
+	if (ferror(stream))
+	{
+		fprintf(stderr, " *ERROR: Problem with output stream of file [%s]!  Skipping...\n", InFilename);
+		if (stream != stdout)
+			fclose(stream);
+		return -1;
+	}
+
+	// loop until end of file
+	while(true)
+	{
+		bzf = BZ2_bzReadOpen(&bzerr, zStream, Verbosity, 0, unused, nUnused);
+		if (bzf == NULL || bzerr != BZ_OK)
+		{
+			ret = testBZ2ErrorHandling(bzerr, bzf, streamNo);
+			if (zStream != stdin)
+				fclose(zStream);
+			if (stream != stdout)
+				fclose(stream);
+			return ret;
+		}
+
+		streamNo++;
+		
+		while (bzerr == BZ_OK)
+		{
+			nread = BZ2_bzRead(&bzerr, bzf, obuf, sizeof(obuf));
+			if (bzerr == BZ_DATA_ERROR_MAGIC)
+			{
+				// try alternate way of reading data
+				if (ForceOverwrite == 1)
+				{
+					rewind(zStream);
+					while (true)
+					{
+						int c = fgetc(zStream);
+						if (c == EOF)
+							break;
+						ungetc(c,zStream);
+					 
+						nread = fread(obuf, sizeof(unsigned char), sizeof(obuf), zStream );
+				      	if (ferror(zStream))
+						{
+							ret = testBZ2ErrorHandling(bzerr, bzf, streamNo);
+							if (zStream != stdin)
+								fclose(zStream);
+							if (stream != stdout)
+								fclose(stream);
+							return ret;
+						}
+						if (nread > 0)
+							fwrite (obuf, sizeof(unsigned char), nread, stream);
+						if (ferror(stream))
+						{
+							ret = testBZ2ErrorHandling(bzerr, bzf, streamNo);
+							if (zStream != stdin)
+								fclose(zStream);
+							if (stream != stdout)
+								fclose(stream);
+							return ret;
+						}
+					}
+					goto closeok;
+				}
+			}
+			if ((bzerr == BZ_OK || bzerr == BZ_STREAM_END) && nread > 0)
+				fwrite(obuf, sizeof(unsigned char), nread, stream );
+			if (ferror(stream))
+			{
+				ret = testBZ2ErrorHandling(bzerr, bzf, streamNo);
+				if (zStream != stdin)
+					fclose(zStream);
+				if (stream != stdout)
+					fclose(stream);
+				return ret;
+			}
+		}
+		if (bzerr != BZ_STREAM_END)
+		{
+			ret = testBZ2ErrorHandling(bzerr, bzf, streamNo);
+			if (zStream != stdin)
+				fclose(zStream);
+			if (stream != stdout)
+				fclose(stream);
+			return ret;
+		}
+
+		BZ2_bzReadGetUnused(&bzerr, bzf, (void**)(&unusedTmp), &nUnused);
+		if (bzerr != BZ_OK)
+		{
+			fprintf(stderr, " *ERROR: Unexpected error. Aborting!\n");
+			exit(3);
+		}
+
+		for (i = 0; i < nUnused; i++)
+			unused[i] = unusedTmp[i];
+
+		BZ2_bzReadClose(&bzerr, bzf);
+		if (bzerr != BZ_OK)
+		{
+			fprintf(stderr, " *ERROR: Unexpected error. Aborting!\n");
+			exit(3);
+		}
+
+		// check to see if we are at the end of the file
+		if (nUnused == 0)
+		{
+			int c = fgetc(zStream);
+			if (c == EOF)
+				break;
+			ungetc(c, zStream);
+		}
+	}
+	
+closeok:
+	// check file stream for errors
+	if (ferror(zStream))
+	{
+		fprintf(stderr, " *ERROR: Problem with intput stream of file [%s]!  Skipping...\n", InFilename);
+		if (zStream != stdin)
+			fclose(zStream);
+		if (stream != stdout)
+			fclose(stream);
+		return -1;
+	}
+	// close file
+	ret = fclose(zStream);
+	if (ret == EOF)
+	{
+		fprintf(stderr, " *ERROR: Problem closing file [%s]!  Skipping...\n", InFilename);
+		return -1;
+	}
+
+	// check file stream for errors
+	if (ferror(stream))
+	{
+		fprintf(stderr, " *ERROR: Problem with output stream of file [%s]!  Skipping...\n", InFilename);
+		if (stream != stdout)
+			fclose(stream);
+		return -1;
+	}
+	ret = fflush(stream);
+	if (ret != 0)
+	{
+		fprintf(stderr, " *ERROR: Problem with output stream of file [%s]!  Skipping...\n", InFilename);
+		if (stream != stdout)
+			fclose(stream);
+		return -1;
+	}
+	if (stream != stdout)
+	{
+		ret = fclose(stream);
+		if (ret == EOF)
+		{
+			fprintf(stderr, " *ERROR: Problem closing file [%s]!  Skipping...\n", OutFilename);
+			return -1;
+		}
+	}
+
+	return 0;
+}
+
+/*
+ * Simulate an unconditional read(), reading in data to fill the
+ * bsize-sized buffer if it can, even if it means calling read() multiple
+ * times. This is needed since pipes and other "special" streams
+ * sometimes don't allow reading of arbitrary sized buffers.
+ */
+ssize_t bufread(int hf, char *buf, size_t bsize)
+{
+	size_t bufr = 0;
+	int ret;
+	int rsize = bsize;
+
+	while (1)
+	{
+		ret = read(hf, buf, rsize);
+
+		if (ret < 0)
+			return ret;
+		if (ret == 0)
+			return bufr;
+
+		bufr += ret;
+		if (bufr == bsize)
+			return bsize;
+		rsize -= ret;
+		buf += ret;
+	}
+}
+
+/*
+ *********************************************************
+ */
+int producer(int hInfile, int blockSize, queue *fifo)
 {
 	char *FileData = NULL;
-	OFF_T bytesLeft = 0;
 	OFF_T inSize = 0;
 	int blockNum = 0;
 	int ret = 0;
 	int pret = -1;
 
-	bytesLeft = fileSize;
+	// We will now totally ignore the fileSize and read the data as it
+	// comes in. Aside from allowing us to process arbitrary streams, it's
+	// also the *right thing to do* in unix environments where data may
+	// be appended to the file as it's processed (e.g. log files).
 
 	// keep going until all the file is processed
-	while (bytesLeft > 0)
+	while (1)
 	{
 		// set buffer size
-		if (bytesLeft > blockSize)
-			inSize = blockSize;
-		else
-			inSize = bytesLeft;
+		inSize = blockSize;
 
 		#ifdef PBZIP_DEBUG
 		fprintf(stderr, " -> Bytes To Read: %"PRIu64" bytes...\n", inSize);
@@ -1064,16 +1329,18 @@ int producer(int hInfile, OFF_T fileSize, int blockSize, queue *fifo)
 		}
 
 		// read file data
-		ret = read(hInfile, (char *) FileData, inSize);
+		ret = bufread(hInfile, (char *) FileData, inSize);
 		#ifdef PBZIP_DEBUG
 		fprintf(stderr, " -> Total Bytes Read: %d bytes...\n\n", ret);
 		#endif
 		if (ret == 0)
 		{
+			// finished reading.
 			pthread_mutex_lock(MemMutex);
 			if (FileData != NULL)
 				delete [] FileData;
 			pthread_mutex_unlock(MemMutex);
+			NumBlocks = blockNum;
 			break;
 		}
 		else if (ret < 0)
@@ -1088,8 +1355,6 @@ int producer(int hInfile, OFF_T fileSize, int blockSize, queue *fifo)
 			return -1;
 		}
 
-		// set bytes left after read
-		bytesLeft -= ret;
 		// check to make sure all the data we expected was read in
 		if (ret != inSize)
 			inSize = ret;
@@ -1118,6 +1383,23 @@ int producer(int hInfile, OFF_T fileSize, int blockSize, queue *fifo)
 		pthread_cond_signal(fifo->notEmpty);
 
 		blockNum++;
+
+		// make sure output buffer is large enough to handle input data
+		if (blockNum > OutputBuffer.size()) 
+		{
+			int newsize = OutputBuffer.size()*2;
+			#ifdef PBZIP_DEBUG
+			fprintf(stderr, "producer:  Resizing OutputBuffer to %d\n", newsize);
+			#endif
+			pthread_mutex_lock(OutMutex);
+			OutputBuffer.resize(newsize);
+			pthread_mutex_unlock(OutMutex);
+			if (OutputBuffer.size() != newsize)
+			{
+				allDone = 1;
+				return -1;
+			}
+		}
 	} // while
 
 	close(hInfile);
@@ -1201,7 +1483,7 @@ void *consumer (void *q)
 		if (pret != 0)
 			fprintf(stderr, "consumer:  *ERROR: pthread_cond_signal error = %d\n", pret);
 		#ifdef PBZIP_DEBUG
-		printf ("consumer: recieved %d.\n", blockNum);
+		fprintf(stderr, "consumer: received %d.\n", blockNum);
 		#endif
 
 		outSize = (int) ((inSize*1.01)+600);
@@ -1446,26 +1728,33 @@ int testCompressedData(char *fileName)
 	unsigned char obuf[5000];
 	unsigned char unused[BZ_MAX_UNUSED];
 	unsigned char *unusedTmp;
-	int bzerr, bzerr_dummy, nread, streamNo;
+	int bzerr, nread, streamNo;
 	int nUnused;
 	int i;
 
 	nUnused = 0;
 	streamNo = 0;
 
-	// open the file for reading
-	zStream = fopen(fileName, "rb");
-	if (zStream == NULL)
+	// see if we are using stdin or not
+	if (strcmp(fileName, "-") != 0) 
 	{
-		fprintf(stderr, " *ERROR: Could not open input file [%s]!  Skipping...\n", fileName);
-		return -1;
+		// open the file for reading
+		zStream = fopen(fileName, "rb");
+		if (zStream == NULL)
+		{
+			fprintf(stderr, " *ERROR: Could not open input file [%s]!  Skipping...\n", fileName);
+			return -1;
+		}
 	}
+	else
+		zStream = stdin;
 
 	// check file stream for errors
 	if (ferror(zStream))
 	{
 		fprintf(stderr, " *ERROR: Problem with stream of file [%s]!  Skipping...\n", fileName);
-		fclose(zStream);
+		if (zStream != stdin)
+			fclose(zStream);
 		return -1;
 	}
 
@@ -1476,7 +1765,8 @@ int testCompressedData(char *fileName)
 		if (bzf == NULL || bzerr != BZ_OK)
 		{
 			ret = testBZ2ErrorHandling(bzerr, bzf, streamNo);
-			fclose(zStream);
+			if (zStream != stdin)
+				fclose(zStream);
 			return ret;
 		}
 
@@ -1488,14 +1778,16 @@ int testCompressedData(char *fileName)
 			if (bzerr == BZ_DATA_ERROR_MAGIC)
 			{
 				ret = testBZ2ErrorHandling(bzerr, bzf, streamNo);
-				fclose(zStream);
+				if (zStream != stdin)
+					fclose(zStream);
 				return ret;
 			}
 		}
 		if (bzerr != BZ_STREAM_END)
 		{
 			ret = testBZ2ErrorHandling(bzerr, bzf, streamNo);
-			fclose(zStream);
+			if (zStream != stdin)
+				fclose(zStream);
 			return ret;
 		}
 
@@ -1531,7 +1823,8 @@ int testCompressedData(char *fileName)
 	if (ferror(zStream))
 	{
 		fprintf(stderr, " *ERROR: Problem with stream of file [%s]!  Skipping...\n", fileName);
-		fclose(zStream);
+		if (zStream != stdin)
+			fclose(zStream);
 		return -1;
 	}
 
@@ -1551,12 +1844,8 @@ int testCompressedData(char *fileName)
  */
 int getFileMetaData(char *fileName)
 {
-	int ret = 0;
-
 	// get the file meta data and store it in the global structure
-	ret = stat(fileName, &fileMetaData);
-
-	return ret;
+	return stat(fileName, &fileMetaData);
 }
 
 /*
@@ -1589,13 +1878,34 @@ int writeFileMetaData(char *fileName)
 	return 0;
 }
 
+/*
+ *********************************************************
+ */
+int detectCPUs()
+{
+	int ncpu;
+	// Autodetect the number of CPUs on a box, if available
+	#if defined(__APPLE__)
+		size_t len = sizeof(ncpu);
+		int mib[2];
+		mib[0] = CTL_HW;
+		mib[1] = HW_NCPU;
+		if (sysctl(mib, 2, &ncpu, &len, 0, 0) < 0 || len != sizeof(ncpu))
+			ncpu = 1;
+	#elif defined(_SC_NPROCESSORS_ONLN)
+		ncpu = sysconf(_SC_NPROCESSORS_ONLN);
+	#endif
+	return ncpu;
+}
+
+
 /*
  *********************************************************
  */
 void banner()
 {
-	fprintf(stderr, "Parallel BZIP2 v1.0.2 - by: Jeff Gilchrist [http://compression.ca]\n");
-	fprintf(stderr, "[July 25, 2007]             (uses libbzip2 by Julian Seward)\n");
+	fprintf(stderr, "Parallel BZIP2 v1.0.3 - by: Jeff Gilchrist [http://compression.ca]\n");
+	fprintf(stderr, "[Oct. 31, 2008]             (uses libbzip2 by Julian Seward)\n");
 
 	return;
 }
@@ -1603,10 +1913,10 @@ void banner()
 /*
  *********************************************************
  */
-void usage(char* progname)
+void usage(char* progname, const char *reason)
 {
 	banner();
-	fprintf(stderr, "\nInvalid command line!  Aborting...\n\n");
+	fprintf(stderr, "\nInvalid command line: %s.  Aborting...\n\n", reason);
 	fprintf(stderr, "Usage: %s [-1 .. -9] [-b#cdfklp#qrtV] <filename> <filename2> <filenameN>\n", progname);
 	fprintf(stderr, " -b#      : where # is the file block size in 100k (default 9 = 900k)\n");
 	fprintf(stderr, " -c       : output to standard out (stdout)\n");
@@ -1618,7 +1928,7 @@ void usage(char* progname)
 #endif
 	fprintf(stderr, " -p#      : where # is the number of processors (default");
 #if defined(_SC_NPROCESSORS_ONLN) || defined(__APPLE__)
-	fprintf(stderr, ": autodetect)\n");
+	fprintf(stderr, ": autodetect [%d])\n", detectCPUs());
 #else
 	fprintf(stderr, " 2)\n");
 #endif
@@ -1629,6 +1939,7 @@ void usage(char* progname)
 	fprintf(stderr, " -1 .. -9 : set BWT block size to 100k .. 900k (default 900k)\n\n");
 	fprintf(stderr, "Example: pbzip2 -b15vk myfile.tar\n");
 	fprintf(stderr, "Example: pbzip2 -p4 -r -5 myfile.tar second*.txt\n");
+	fprintf(stderr, "Example: tar -c directory_to_compress/ | pbzip2 -c > myfile.tar.bz2\n");
 	fprintf(stderr, "Example: pbzip2 -d myfile.tar.bz2\n\n");
 	exit(-1);
 }
@@ -1645,9 +1956,11 @@ int main(int argc, char* argv[])
 	char *progName = NULL;
 	char *progNamePos = NULL;
 	char bz2Header[] = {"BZh91AY&SY"};  // using 900k block size
-	char OutFilename[255];
-	char cmdLineTemp[255];
+	char bz2HeaderZero[] = { 0x42, 0x5A, 0x68, 0x39, 0x17, 0x72, 0x45, 0x38, 0x50, 0x90, 0x00, 0x00, 0x00, 0x00 };
+	char OutFilename[2048];
+	char cmdLineTemp[2048];
 	char tmpBuff[50];
+	char stdinFile[2] = {"-"};
 	struct timeval tvStartTime;
 	struct timeval tvStopTime;
 	#ifndef WIN32
@@ -1663,7 +1976,6 @@ int main(int argc, char* argv[])
 	LARGE_INTEGER fileSize_temp;
 	HANDLE hInfile_temp;
 	#endif
-	struct stat statbuf;
 	double timeCalc = 0.0;
 	double timeStart = 0.0;
 	double timeStop = 0.0;
@@ -1671,6 +1983,7 @@ int main(int argc, char* argv[])
 	size_t size;
 	int cmdLineTempCount = 0;
 	int readEntireFile = 0;
+	int zeroByteFile = 0;
 	int hInfile = -1;
 	int hOutfile = -1;
 	int numBlocks = 0;
@@ -1698,7 +2011,7 @@ int main(int argc, char* argv[])
 	// check to make sure proper command line is given
 	if (argc < 2)
 	{
-		usage(argv[0]);
+		usage(argv[0], "Not enough arguments");
 	}
 
 	// get program name to determine if decompress mode should be used
@@ -1709,7 +2022,13 @@ int main(int argc, char* argv[])
 			progName = progNamePos + 1;
 	}
 	if ((strstr(progName, "unzip") != 0) || (strstr(progName, "UNZIP") != 0))
+	{
 		decompress = 1;
+	}
+	if ((strstr(progName, "zcat") != 0) || (strstr(progName, "ZCAT") != 0))
+	{
+		decompress = OutputStdOut = keep = 1; 
+	}
 	
 	FileListCount = 0;
 	FileList = new char *[argc];
@@ -1719,17 +2038,8 @@ int main(int argc, char* argv[])
 		return 1;
 	}
 
-	// Autodetect the number of CPUs on a box, if available
-	#if defined(__APPLE__)
-		size_t len = sizeof(numCPU);
-		int mib[2];
-		mib[0] = CTL_HW;
-		mib[1] = HW_NCPU;
-		if (sysctl(mib, 2, &numCPU, &len, 0, 0) < 0 || len != sizeof(numCPU))
-			numCPU = 1;
-	#elif defined(_SC_NPROCESSORS_ONLN)
-		numCPU = sysconf(_SC_NPROCESSORS_ONLN);
-	#endif
+	numCPU = detectCPUs();
+
 	#ifndef WIN32
 	numCPUtotal = numCPU;
 	#endif
@@ -1739,6 +2049,13 @@ int main(int argc, char* argv[])
 	{
 		if (argv[i][0] == '-')
 		{
+			if (argv[i][1] == '\0') 
+			{
+				// support "-" as a filename
+				FileList[FileListCount] = argv[i];
+				FileListCount++;
+				continue;
+			}
 			#ifdef PBZIP_DEBUG
 			fprintf(stderr, "argv[%d]: %s   Len: %d\n", i, argv[i], strlen(argv[i]));
 			#endif
@@ -1757,7 +2074,7 @@ int main(int argc, char* argv[])
 						cmdLineTempCount++;
 					}
 					if (cmdLineTempCount == 0)
-						usage(argv[0]);
+						usage(argv[0], "Cannot parse -p argument");
 					strncpy(cmdLineTemp, argv[i]+j+1, cmdLineTempCount);
 					numCPU = atoi(cmdLineTemp);
 					if (numCPU > 4096)
@@ -1785,7 +2102,7 @@ int main(int argc, char* argv[])
 						cmdLineTempCount++;
 					}
 					if (cmdLineTempCount == 0)
-						usage(argv[0]);
+						usage(argv[0], "Cannot parse file block size");
 					strncpy(cmdLineTemp, argv[i]+j+1, cmdLineTempCount);
 					blockSize = atoi(cmdLineTemp)*100000;
 					if ((blockSize < 100000) || (blockSize > 1000000000))
@@ -1800,7 +2117,7 @@ int main(int argc, char* argv[])
 					break;
 				case 'd': decompress = 1; break;
 				case 'c': OutputStdOut = 1; keep = 1; break;
-				case 'f': force = 1; break;
+				case 'f': force = 1; ForceOverwrite = 1; break;
 				case 'k': keep = 1; break;
 				#ifndef PBZIP_NO_LOADAVG
 				case 'l': useLoadAverage = 1; break;
@@ -1830,7 +2147,16 @@ int main(int argc, char* argv[])
 	} /* for */
 
 	if (FileListCount == 0)
-		usage(argv[0]);
+	{
+		if ((OutputStdOut == 1) || (testFile == 1))
+		{
+			// expecting data from stdin
+			FileList[FileListCount] = stdinFile;
+			FileListCount++;
+		}
+		else
+			usage(argv[0], "Not enough files given");
+	}
 
 	if (QuietMode != 1)
 	{
@@ -1900,7 +2226,7 @@ int main(int argc, char* argv[])
 			fprintf(stderr, "\n         # CPUs: %d\n", numCPU);
 			#ifndef PBZIP_NO_LOADAVG
 			if (useLoadAverage == 1)
-				fprintf(stderr, "   Load Average: %'.2f\n", loadAverage);
+				fprintf(stderr, "   Load Average: %.2f\n", loadAverage);
 			#endif
 			if (decompress != 1)
 			{
@@ -1945,6 +2271,7 @@ int main(int argc, char* argv[])
 	for (fileLoop=0; fileLoop < FileListCount; fileLoop++)
 	{
 		allDone = 0;
+		fileSize = 0;
 
 		// set input filename
 		InFilename = FileList[fileLoop];
@@ -1955,7 +2282,10 @@ int main(int argc, char* argv[])
 			if (QuietMode != 1)
 			{
 				fprintf(stderr, "      File #: %d of %d\n", fileLoop+1, FileListCount);
-				fprintf(stderr, "     Testing: %s\n", InFilename);
+				if (strcmp(InFilename, "-") != 0) 
+					fprintf(stderr, "     Testing: %s\n", InFilename);
+				else
+					fprintf(stderr, "     Testing: <stdin>\n");
 			}
 			ret = testCompressedData(InFilename);
 			if (ret > 0)
@@ -1974,8 +2304,8 @@ int main(int argc, char* argv[])
 		}
 
 		// set ouput filename
-		strncpy(OutFilename, FileList[fileLoop], 250);
-		if (decompress == 1)
+		strncpy(OutFilename, FileList[fileLoop], 2040);
+		if ((decompress == 1) && (strcmp(InFilename, "-") != 0))
 		{
 			// check if input file is a valid .bz2 compressed file
 			hInfile = open(InFilename, O_RDONLY | O_BINARY);
@@ -2010,10 +2340,17 @@ int main(int argc, char* argv[])
 				// skip 4th char which differs depending on BWT block size used
 				if (memstr(tmpBuff+4, size-4, bz2Header+4, strlen(bz2Header)-4) == NULL)
 				{
-					fprintf(stderr, " *ERROR: File [%s] is NOT a valid bzip2!  Skipping...\n", InFilename);
-					fprintf(stderr, "-------------------------------------------\n");
-					errLevel = 1;
-					continue;
+					// check to see if this is a special 0 byte file
+					if (memstr(tmpBuff+4, size-4, bz2HeaderZero+4, strlen(bz2Header)-4) == NULL)
+					{
+						fprintf(stderr, " *ERROR: File [%s] is NOT a valid bzip2!  Skipping...\n", InFilename);
+						fprintf(stderr, "-------------------------------------------\n");
+						errLevel = 1;
+						continue;
+					}
+					#ifdef PBZIP_DEBUG
+					fprintf(stderr, "** ZERO byte compressed file detected\n");
+					#endif
 				}
 				// set block size for decompression
 				if ((tmpBuff[3] >= '1') && (tmpBuff[3] <= '9'))
@@ -2038,7 +2375,7 @@ int main(int argc, char* argv[])
 				// add .out extension so we don't overwrite original file
 				strcat(OutFilename, ".out");
 			}
-		}
+		} // decompress == 1
 		else
 		{
 			// check input file to make sure its not already a .bz2 file
@@ -2056,58 +2393,75 @@ int main(int argc, char* argv[])
 		sigInFilename = InFilename;
 		sigOutFilename = OutFilename;
 
-		// read file for compression
-		hInfile = open(InFilename, O_RDONLY | O_BINARY);
-		// check to see if file exists before processing
-		if (hInfile == -1)
+		if (strcmp(InFilename, "-") != 0) 
 		{
-			fprintf(stderr, " *ERROR: File [%s] NOT found!  Skipping...\n", InFilename);
-			fprintf(stderr, "-------------------------------------------\n");
-			errLevel = 1;
-			continue;
-		}
+			struct stat statbuf;
+			// read file for compression
+			hInfile = open(InFilename, O_RDONLY | O_BINARY);
+			// check to see if file exists before processing
+			if (hInfile == -1)
+			{
+				fprintf(stderr, " *ERROR: File [%s] NOT found!  Skipping...\n", InFilename);
+				fprintf(stderr, "-------------------------------------------\n");
+				errLevel = 1;
+				continue;
+			}
 
-		// get some information about the file
-		fstat(hInfile, &statbuf);
-		// check to make input is not a directory
-		if (S_ISDIR(statbuf.st_mode))
-		{
-			fprintf(stderr, " *ERROR: File [%s] is a directory!  Skipping...\n", InFilename);
-			fprintf(stderr, "-------------------------------------------\n");
-			errLevel = 1;
-			continue;
-		}
-		// check to make sure input is a regular file
-		if (!S_ISREG(statbuf.st_mode))
-		{
-			fprintf(stderr, " *ERROR: File [%s] is not a regular file!  Skipping...\n", InFilename);
-			fprintf(stderr, "-------------------------------------------\n");
-			errLevel = 1;
-			continue;
-		}
-		// get size of file
-		#ifndef WIN32
-		fileSize = statbuf.st_size;
-		#else
-		fileSize_temp.LowPart = GetFileSize((HANDLE)_get_osfhandle(hInfile), (unsigned long *)&fileSize_temp.HighPart);
-		fileSize = fileSize_temp.QuadPart;
-		#endif
-		// don't process a 0 byte file
-		if (fileSize == 0)
-		{
-			fprintf(stderr, " *ERROR: File is of size 0 [%s]!  Skipping...\n", InFilename);
-			fprintf(stderr, "-------------------------------------------\n");
-			errLevel = 1;
-			continue;
-		}
+			// get some information about the file
+			fstat(hInfile, &statbuf);
+			// check to make input is not a directory
+			if (S_ISDIR(statbuf.st_mode))
+			{
+				fprintf(stderr, " *ERROR: File [%s] is a directory!  Skipping...\n", InFilename);
+				fprintf(stderr, "-------------------------------------------\n");
+				errLevel = 1;
+				continue;
+			}
+			// check to make sure input is a regular file
+			if (!S_ISREG(statbuf.st_mode))
+			{
+				fprintf(stderr, " *ERROR: File [%s] is not a regular file!  Skipping...\n", InFilename);
+				fprintf(stderr, "-------------------------------------------\n");
+				errLevel = 1;
+				continue;
+			}
+			// get size of file
+			#ifndef WIN32
+			fileSize = statbuf.st_size;
+			#else
+			fileSize_temp.LowPart = GetFileSize((HANDLE)_get_osfhandle(hInfile), (unsigned long *)&fileSize_temp.HighPart);
+			fileSize = fileSize_temp.QuadPart;
+			#endif
+			// don't process a 0 byte file
+			if (fileSize == 0)
+			{
+				if (decompress == 1)
+				{
+					fprintf(stderr, " *ERROR: File is of size 0 [%s]!  Skipping...\n", InFilename);
+					fprintf(stderr, "-------------------------------------------\n");
+					errLevel = 1;
+					continue;
+				}
+				
+				// make sure we handle zero byte files specially
+				zeroByteFile = 1;
+			}
+			else
+				zeroByteFile = 0;
 
-		// get file meta data to write to output file
-		if (getFileMetaData(InFilename) != 0)
+			// get file meta data to write to output file
+			if (getFileMetaData(InFilename) != 0)
+			{
+				fprintf(stderr, " *ERROR: Could not get file meta data from [%s]!  Skipping...\n", InFilename);
+				fprintf(stderr, "-------------------------------------------\n");
+				errLevel = 1;
+				continue;
+			}
+		}
+		else
 		{
-			fprintf(stderr, " *ERROR: Could not get file meta data from [%s]!  Skipping...\n", InFilename);
-			fprintf(stderr, "-------------------------------------------\n");
-			errLevel = 1;
-			continue;
+			hInfile = 0;	// stdin
+			fileSize = -1;	// fake it
 		}
 
 		// check to see if output file exists
@@ -2128,15 +2482,24 @@ int main(int argc, char* argv[])
 
 		if (readEntireFile == 1)
 		{
-			// determine block size to try and spread data equally over # CPUs
-			blockSize = fileSize / numCPU;
+			if (hInfile == 0) 
+			{
+				if (QuietMode != 1)
+					fprintf(stderr, " *Warning: Ignoring -r switch since input is stdin.\n");
+			}
+			else
+			{
+				// determine block size to try and spread data equally over # CPUs
+				blockSize = fileSize / numCPU;
+			}
 		}
 
 		// display per file settings
 		if (QuietMode != 1)
 		{
 			fprintf(stderr, "         File #: %d of %d\n", fileLoop+1, FileListCount);
-			fprintf(stderr, "     Input Name: %s\n", InFilename);
+			fprintf(stderr, "     Input Name: %s\n", hInfile != 0 ? InFilename : "<stdin>");
+
 			if (OutputStdOut == 0)
 				fprintf(stderr, "    Output Name: %s\n\n", OutFilename);
 			else
@@ -2144,24 +2507,76 @@ int main(int argc, char* argv[])
 
 			if (decompress == 1)
 				fprintf(stderr, " BWT Block Size: %c00k\n", BWTblockSizeChar);
-			fprintf(stderr, "     Input Size: %"PRIu64" bytes\n", (unsigned long long)fileSize);
+			if (strcmp(InFilename, "-") != 0) 
+				fprintf(stderr, "     Input Size: %"PRIu64" bytes\n", (unsigned long long)fileSize);
 		}
 
 		if (decompress == 1)
 		{
 			numBlocks = 0;
-			noThreads = 0;
-		}
-		else
-		{
-			// calculate the # of blocks of data
-			numBlocks = (fileSize + blockSize - 1) / blockSize;
-			// Do not use threads for small files where we only have 1 block to process
-			// or if we only have 1 CPU
-			if ((numBlocks == 1) || (numCPU == 1))
+			// Do not use threads if we only have 1 CPU or small files
+			if ((numCPU == 1) || (fileSize < 1000000))
 				noThreads = 1;
 			else
 				noThreads = 0;
+			// for now use no threads method for uncompressing from stdin
+			if (strcmp(InFilename, "-") == 0)
+				noThreads = 1;
+		}
+		else
+		{
+			if (fileSize > 0) 
+			{
+				// calculate the # of blocks of data
+				numBlocks = (fileSize + blockSize - 1) / blockSize;
+				// Do not use threads for small files where we only have 1 block to process
+				// or if we only have 1 CPU
+				if ((numBlocks == 1) || (numCPU == 1))
+					noThreads = 1;
+				else
+					noThreads = 0;
+			} 
+			else 
+			{
+				// Simulate a "big" number of buffers. Will need to resize it later
+				numBlocks = 10000;
+			}
+			
+			// write special compressed data for special 0 byte input file case
+			if (zeroByteFile == 1)
+			{
+				hOutfile = 1;
+				// write to file instead of stdout
+				if (OutputStdOut == 0)
+				{
+					hOutfile = open(OutFilename, O_RDWR | O_CREAT | O_TRUNC | O_BINARY, FILE_MODE);
+					// check to see if file creation was successful
+					if (hOutfile == -1)
+					{
+						fprintf(stderr, " *ERROR: Could not create output file [%s]!\n", OutFilename);
+						close(hOutfile);
+						errLevel = 1;
+						continue;
+					}
+				}
+				// write data to the output file
+				ret = write(hOutfile, bz2HeaderZero, sizeof(bz2HeaderZero));
+				if (OutputStdOut == 0)
+					close(hOutfile);
+				if (ret != sizeof(bz2HeaderZero))
+				{
+					fprintf(stderr, " *ERROR: Could not write to file!  Skipping...\n");
+					fprintf(stderr, "-------------------------------------------\n");
+					errLevel = 1;
+					continue;
+				}
+				if (QuietMode != 1)
+				{
+					fprintf(stderr, "    Output Size: %"PRIu64" bytes\n", (unsigned long long)sizeof(bz2HeaderZero));
+					fprintf(stderr, "-------------------------------------------\n");
+				}
+				continue;
+			}
 		}
 		#ifdef PBZIP_DEBUG
 		fprintf(stderr, "# Blocks: %d\n", numBlocks);
@@ -2186,31 +2601,53 @@ int main(int argc, char* argv[])
 
 		if (decompress == 1)
 		{
-			// do decompression
-			NumBufferedBlocks = 0;
-			if (QuietMode != 1)
-				fprintf(stderr, "Decompressing data...\n");
-			for (i=0; i < numCPU; i++)
+			// use multi-threaded code
+			if (noThreads == 0)
 			{
-				ret = pthread_create(&fifo->consumers[i], NULL, consumer_decompress, fifo);
+				// do decompression
+				NumBufferedBlocks = 0;
+				if (QuietMode != 1)
+					fprintf(stderr, "Decompressing data...\n");
+				for (i=0; i < numCPU; i++)
+				{
+					ret = pthread_create(&fifo->consumers[i], NULL, consumer_decompress, fifo);
+					if (ret != 0)
+					{
+						fprintf(stderr, " *ERROR: Not enough resources to create consumer thread #%d (code = %d)  Aborting...\n", i, ret);
+						return 1;
+					}
+				}
+
+				ret = pthread_create(&output, NULL, fileWriter, OutFilename);
 				if (ret != 0)
 				{
-					fprintf(stderr, " *ERROR: Not enough resources to create consumer thread #%d (code = %d)  Aborting...\n", i, ret);
+					fprintf(stderr, " *ERROR: Not enough resources to create fileWriter thread (code = %d)  Aborting...\n", ret);
 					return 1;
 				}
-			}
 
-			ret = pthread_create(&output, NULL, fileWriter, OutFilename);
-			if (ret != 0)
+				// start reading in data for decompression
+				ret = producer_decompress(hInfile, fileSize, fifo);
+				if (ret == -99)
+				{
+					// only 1 block detected, use single threaded code to decompress
+					noThreads = 1;
+				}
+				else if (ret != 0)
+					errLevel = 1;
+			}
+			
+			// use single threaded code
+			if (noThreads == 1)
 			{
-				fprintf(stderr, " *ERROR: Not enough resources to create fileWriter thread (code = %d)  Aborting...\n", ret);
-				return 1;
-			}
+				if (QuietMode != 1)
+					fprintf(stderr, "Decompressing data (no threads)...\n");
 
-			// start reading in data for decompression
-			ret = producer_decompress(hInfile, fileSize, fifo);
-			if (ret != 0)
-				errLevel = 1;
+				if (hInfile > 0)
+					close(hInfile);
+				ret = directdecompress(InFilename, OutFilename);
+				if (ret != 0)
+					errLevel = 1;
+			}
 		}
 		else
 		{
@@ -2240,7 +2677,7 @@ int main(int argc, char* argv[])
 				}
 	
 				// start reading in data for compression
-				ret = producer(hInfile, fileSize, blockSize, fifo);
+				ret = producer(hInfile, blockSize, fifo);
 				if (ret != 0)
 					errLevel = 1;
 			}
@@ -2268,21 +2705,9 @@ int main(int argc, char* argv[])
 
 		if (OutputStdOut == 0)
 		{
-			statbuf.st_size = 0;
-			// check to see if output file is larger than 0 bytes
-			stat(OutFilename, &statbuf);
-			// get size of file
-			if (statbuf.st_size == 0)
-			{
-				fprintf(stderr, " *ERROR: Output file is size 0 bytes!  Deleting...\n");
-				remove(OutFilename);
-			}
-			else
-			{
-				// write store file meta data to output file
-				if (writeFileMetaData(OutFilename) != 0)
-					fprintf(stderr, " *ERROR: Could not write file meta data to [%s]!\n", InFilename);
-			}
+			// write store file meta data to output file
+			if (writeFileMetaData(OutFilename) != 0)
+				fprintf(stderr, " *ERROR: Could not write file meta data to [%s]!\n", InFilename);
 		}
 
 		// finished processing file
@@ -2292,6 +2717,7 @@ int main(int argc, char* argv[])
 		// remove input file unless requested not to by user
 		if (keep != 1)
 		{
+			struct stat statbuf;
 			if (OutputStdOut == 0)
 			{
 				// only remove input file if output file exists
